#include "bas_niv.h"

/* Lit le noeud correspondant a 'num_noeud' (de numero de noeud num_noeud.num
   sur le disque numero num_noeud.dev) et remplit 'buf'.
   Retourne 0 en cas de succes, -1 sinon.
   'buf' doit etre un pointeur sur une zone memoire alloue
   et celle-ci n'est pas automatiquement liberee en cas d'echec. 
   Ne genere que des erreurs de type ERREUR_INCONNUE (ie pour en savoir plus
   voir la variable errno directement) */ 
int read_node (noeud_p buf, n_noeud_s num_noeud) {
  struct flock *verrou;
  int num_bloc;
  noeud_p buf_bloc;
  int desc_hd;
  DEV *dev_p;

  dev_p = mount_table + num_noeud.dev;
  desc_hd = dev_p->desc;
  num_bloc = NOEUD2BLOC(num_noeud.num);
  /* Verification du disque ou on veut lire */
  if (dev_p->dev != num_noeud.dev) {
    fprintf(stderr,"Lecture ds un disque inconnu: %d\n",num_noeud.dev);
    num_err = DISQUE_INCONNU;
    return -1;
  }
  /* Verification du noeud qu'on veut lire */
  if (num_noeud.num < 0  || num_noeud.num >= NOEUDS_PAR_BLOC*TAILLE_ZONE_NOEUDS) {
    fprintf(stderr,"Lecture d'un noeud inexistant: %d\n",num_noeud.num);
    num_err = MAUVAISE_POSITION;
    return -1;
  }
  /* Pose d'un verrou sur le bloc a lire */
  verrou = (struct flock*) malloc(sizeof(struct flock));
  verrou->l_type = F_RDLCK;
  verrou->l_whence = SEEK_SET;
  verrou->l_start = (off_t) num_bloc*TAILLE_BLOC;
  verrou->l_len = (off_t) TAILLE_BLOC;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Lecture du bloc */
  if (lseek(desc_hd , (off_t)num_bloc*TAILLE_BLOC , SEEK_SET) == -1) {
    perror("lseek");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
   return -1;
  }
  buf_bloc = (noeud_p) malloc(TAILLE_BLOC);
  if (read(desc_hd,buf_bloc,TAILLE_BLOC) != TAILLE_BLOC) {
    perror("read");
    free(verrou);
    free(buf_bloc);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  *buf = *(buf_bloc + NOEUDINBLOC(num_noeud.num));
  /* Deverrouillage */
  verrou->l_type = F_UNLCK;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    free(buf_bloc);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }

  free(buf_bloc);
  free(verrou);
  return 0;
}

/* Ecrit le contenu de 'buf' dans le noeud correspondant a 'num_noeud' (de numero
   de noeud num_noeud.num sur le disque numero num_noeud.dev).
   Retourne 0 en cas de succes, -1 sinon.
   Ne genere que des erreurs de type ERREUR_INCONNUE (ie pour en savoir plus
   voir la variable errno directement) */ 
int write_node (noeud_p buf, n_noeud_s num_noeud) {
  struct flock *verrou;
  int num_bloc;
  noeud_p buf_bloc,np;
  int desc_hd;
  DEV *dev_p;

  dev_p = mount_table + num_noeud.dev;
  desc_hd = dev_p->desc;
  num_bloc = NOEUD2BLOC(num_noeud.num);
  /* Verification du disque ou on veut ecrire */
  if (dev_p->dev != num_noeud.dev) {
    fprintf(stderr,"Ecriture ds un disque inconnu: %d\n",num_noeud.dev);
    num_err = DISQUE_INCONNU;
    return -1;
  }
  /* Verification du noeud qu'on veut ecrire */
  if (num_noeud.num < 0  || num_noeud.num >= NOEUDS_PAR_BLOC*TAILLE_ZONE_NOEUDS) {
    fprintf(stderr,"Ecriture sur un noeud inexistant: %d\n",num_noeud.num);
    num_err = MAUVAISE_POSITION;
    return -1;
  }
  /* Pose d'un verrou sur le bloc a lire */
  verrou = (struct flock*) malloc(sizeof(struct flock));
  verrou->l_type = F_WRLCK;
  verrou->l_whence = SEEK_SET;
  verrou->l_start = (off_t) num_bloc*TAILLE_BLOC;
  verrou->l_len = (off_t) TAILLE_BLOC;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Lecture du bloc */
  if (lseek(desc_hd , (off_t)num_bloc*TAILLE_BLOC , SEEK_SET) == -1) {
    perror("lseek");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  buf_bloc = (noeud_p) malloc(TAILLE_BLOC);
  if (read(desc_hd,buf_bloc,TAILLE_BLOC) != TAILLE_BLOC) {
    perror("read");
    free(verrou);
    free(buf_bloc);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Ecriture du noeud */
  np = buf_bloc + NOEUDINBLOC(num_noeud.num);
  *np = *buf;
  if (lseek(desc_hd , (off_t)num_bloc*TAILLE_BLOC , SEEK_SET) == -1) {
    perror("lseek");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
   return -1;
  }
  if (write(desc_hd,buf_bloc,TAILLE_BLOC) != TAILLE_BLOC) {
    perror("write");
    free(verrou);
    free(buf_bloc);
    num_err = ERREUR_INCONNUE; 
   return -1;
  }
  /* Deverrouillage */
  verrou->l_type = F_UNLCK;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    free(buf_bloc);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }

  free(buf_bloc);
  free(verrou);
  return 0;  
}

/* Lit le bloc correspondant a 'num_bloc' (de numero de bloc num_bloc.num
   sur le disque numero numbloc_.dev) et remplit 'buf'.
   Retourne 0 en cas de succes, -1 sinon.
   'buf' doit etre un pointeur sur une zone memoire alloue
   et celle-ci n'est pas automatiquement liberee en cas d'echec. 
   Ne genere que des erreurs de type ERREUR_INCONNUE (ie pour en savoir plus
   voir la variable errno directement) */ 
int read_bloc (void* buf, n_bloc_s num_bloc) {
  struct flock *verrou;
  int desc_hd;
  DEV *dev_p;

  dev_p = mount_table + num_bloc.dev;
  desc_hd = dev_p->desc;
  /* Verification du disque ou on veut lire */
  if (dev_p->dev != num_bloc.dev) {
    fprintf(stderr,"Lecture ds un disque inconnu: %d\n",num_bloc.dev);
    num_err = DISQUE_INCONNU;
    return -1;
  }
  /* Verification du bloc qu'on veut lire */
  if (num_bloc.num < 0  || num_bloc.num >= TAILLE_ZONE_DONNEES) {
    fprintf(stderr,"Lecture d'un bloc inexistant: %d\n",num_bloc.num);
    num_err = MAUVAISE_POSITION;
    return -1;
  }
  /* Pose d'un verrou sur le bloc a lire */
  verrou = (struct flock*) malloc(sizeof(struct flock));
  verrou->l_type = F_RDLCK;
  verrou->l_whence = SEEK_SET;
  verrou->l_start = (off_t) DEBUTBLOC(num_bloc.num);
  verrou->l_len = (off_t) TAILLE_BLOC;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Lecture du bloc */
  if (lseek(desc_hd , (off_t) DEBUTBLOC(num_bloc.num) , SEEK_SET) == -1) {
    perror("lseek");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  if (read(desc_hd,buf,TAILLE_BLOC) != TAILLE_BLOC) {
    perror("read");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Deverrouillage */
  verrou->l_type = F_UNLCK;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }

  free(verrou);
  return 0;
}

/* Ecrit le contenu de 'buf' dans le bloc correspondant a 'num_bloc' 
   (de numero de bloc num_bloc.num sur le disque numero numbloc_.dev).
   Retourne 0 en cas de succes, -1 sinon.
   Ne genere que des erreurs de type ERREUR_INCONNUE (ie pour en savoir plus
   voir la variable errno directement) */ 
int write_bloc (void* buf,n_bloc_s num_bloc) {
  struct flock *verrou;
  int desc_hd;
  DEV *dev_p;

  dev_p = mount_table + num_bloc.dev;
  desc_hd = dev_p->desc;
  /* Verification du disque ou on veut ecrire */
  if (dev_p->dev != num_bloc.dev) {
    fprintf(stderr,"Ecriture ds un disque inconnu: %d\n",num_bloc.dev);
    num_err = DISQUE_INCONNU;
    return -1;
  }
  /* Verification du bloc qu'on veut ecrire */
  if (num_bloc.num < 0  || num_bloc.num >= TAILLE_ZONE_DONNEES) {
    fprintf(stderr,"Ecriture sur un bloc inexistant: %d\n",num_bloc.num);
    num_err = MAUVAISE_POSITION;
    return -1;
  }
  /* Pose d'un verrou sur le bloc a ecrire */
  verrou = (struct flock*) malloc(sizeof(struct flock));
  verrou->l_type = F_WRLCK;
  verrou->l_whence = SEEK_SET;
  verrou->l_start = (off_t) DEBUTBLOC(num_bloc.num);
  verrou->l_len = (off_t) TAILLE_BLOC;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Ecriture du bloc */
  if (lseek(desc_hd , (off_t) DEBUTBLOC(num_bloc.num) , SEEK_SET) == -1) {
    perror("lseek");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  if (write(desc_hd,buf,TAILLE_BLOC) != TAILLE_BLOC) {
    perror("write");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Deverrouillage */
  verrou->l_type = F_UNLCK;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }

  free(verrou);
  return 0;  
}

/* Lit le 1er bloc du disque et remplit la table des noeuds libres (ie 
   'free_nodes_table' et 'cur_p_node') du disque de numero 'dev'. 
   Retourne 0 en cas de succes, -1 sinon.
   Ne genere que des erreurs de type ERREUR_INCONNUE (ie pour en savoir plus
   voir la variable errno directement). */
int read_free_nodes_table (short dev) {
  struct flock *verrou;
  void* buf_bloc;
  int desc_hd;
  DEV *dev_p;

  dev_p = mount_table + dev;
  desc_hd = dev_p->desc;
  /* Verification du disque ou on veut lire la table */
  if (dev_p->dev != dev) {
    fprintf(stderr,"Lecture ds un disque inconnu: %d\n",dev);
    num_err = DISQUE_INCONNU;
    return -1;
  }
  /* Pose d'un verrou sur le bloc a lire */
  verrou = (struct flock*) malloc(sizeof(struct flock));
  verrou->l_type = F_RDLCK;
  verrou->l_whence = SEEK_SET;
  verrou->l_start = (off_t) 0;
  verrou->l_len = (off_t) TAILLE_BLOC;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Lecture du bloc */
  if (lseek(desc_hd , (off_t) 0 , SEEK_SET) == -1) {
    perror("lseek");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  buf_bloc = malloc(TAILLE_BLOC);
  if (read(desc_hd,buf_bloc,TAILLE_BLOC) != TAILLE_BLOC) {
    perror("read");
    free(verrou);
    free(buf_bloc);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Table des noeuds libres */
  dev_p->free_nodes_table = (short*) realloc(buf_bloc,sizeof(short)*(MAX_FREE_NODES+1));
  /* Pointeur courant */
  dev_p->cur_p_node = dev_p->free_nodes_table[MAX_FREE_NODES];
  /* Deverrouillage */
  verrou->l_type = F_UNLCK;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    free(dev_p->free_nodes_table);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }

  free(verrou);
  return 0;    
}

/* Ecrit toute la table de noeuds libres a partir des variables 
   'free_nodes_table' et 'cur_p_node' du disque 'dev'. 
   Retourne 0 en cas de succes, -1 sinon.
   Ne genere que des erreurs de type ERREUR_INCONNUE (ie pour en savoir plus
   voir la variable errno directement). */
int write_free_nodes_table (short dev) {
  struct flock *verrou;
  void* buf_bloc;
  short *buf_tab;
  int *int_p;
  int desc_hd,i;
  DEV *dev_p;

  dev_p = mount_table + dev;
  desc_hd = dev_p->desc;
  /* Verification du disque ou on veut ecrire la table */
  if (dev_p->dev != dev) {
    fprintf(stderr,"Ecriture sur un disque inconnu: %d\n",dev);
    num_err = DISQUE_INCONNU;
    return -1;
  }
  /* Pose d'un verrou sur le bloc a ecrire */
  verrou = (struct flock*) malloc(sizeof(struct flock));
  verrou->l_type = F_WRLCK;
  verrou->l_whence = SEEK_SET;
  verrou->l_start = (off_t) 0;
  verrou->l_len = (off_t) TAILLE_BLOC;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Table des noeuds libres */
  buf_tab = (short*) malloc(sizeof(short)*(MAX_FREE_NODES+1));
  for (i=0 ; i<MAX_FREE_NODES ; i++) {
    buf_tab[i] = dev_p->free_nodes_table[i];
  }
  /* Pointeur courant */
  buf_tab[MAX_FREE_NODES] = dev_p->cur_p_node;
  /* Caracteristique du disque */
  buf_bloc = realloc(buf_tab,TAILLE_BLOC);
  int_p = buf_bloc + (sizeof(short)*(MAX_FREE_NODES+1));
  *int_p = TAILLE_BLOC;
  int_p ++;
  *int_p = TAILLE_ZONE_DESCRIPTION;
  int_p ++;
  *int_p = TAILLE_ZONE_NOEUDS;
  int_p ++;
  *int_p = TAILLE_ZONE_DONNEES;
  /* Ecriture du bloc */
  if (lseek(desc_hd , (off_t)0 , SEEK_SET) == -1) {
    perror("lseek");
    free(verrou);
    free(buf_bloc);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  if (write(desc_hd,buf_bloc,TAILLE_BLOC) != TAILLE_BLOC) {
    perror("write");
    free(verrou);
    free(buf_bloc);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Deverrouillage */
  verrou->l_type = F_UNLCK;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    free(buf_bloc);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }

  free(verrou);
  free(buf_bloc);
  return 0;    
}

/* Lit le 2eme bloc du disque et remplit la table des blocs libres (ie 
   'free_blocs_table' et 'cur_p_bloc') du disque de numero 'dev'.
   Retourne 0 en cas de succes, -1 sinon.
   Ne genere que des erreurs de type ERREUR_INCONNUE (ie pour en savoir plus
   voir la variable errno directement). */
int read_free_blocs_table (short dev) {
  struct flock *verrou;
  int desc_hd;
  DEV *dev_p;

  dev_p = mount_table + dev;
  desc_hd = dev_p->desc;
  /* Verification du disque ou on veut lire la table */
  if (dev_p->dev != dev) {
    fprintf(stderr,"Lecture ds un disque inconnu: %d\n",dev);
    num_err = DISQUE_INCONNU;
    return -1;
  }
  /* Pose d'un verrou sur le bloc a lire */
  verrou = (struct flock*) malloc(sizeof(struct flock));
  verrou->l_type = F_RDLCK;
  verrou->l_whence = SEEK_SET;
  verrou->l_start = (off_t) TAILLE_BLOC;
  verrou->l_len = (off_t) TAILLE_BLOC;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Lecture du bloc */
  if (lseek(desc_hd , (off_t)TAILLE_BLOC , SEEK_SET) == -1) {
    perror("lseek");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  if (read(desc_hd,dev_p->free_blocs_table,TAILLE_BLOC) != TAILLE_BLOC) {
    perror("read");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Lecture du pointeur */
  dev_p->cur_p_bloc = dev_p->free_blocs_table[MAX_FREE_BLOCS];
  /* Deverrouillage */
  verrou->l_type = F_UNLCK;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }

  free(verrou);
  return 0;  
}

/* Ecrit toute la table des blocs libres (sur le 2eme bloc du disque) a partir
   des variables 'free_blocs_table' et 'cur_p_bloc' du disque de numero 'dev'. 
   Retourne 0 en cas de succes, -1 sinon.
   Ne genere que des erreurs de type ERREUR_INCONNUE (ie pour en savoir plus
   voir la variable errno directement). */
int write_free_blocs_table (short dev) {
  struct flock *verrou;
  int desc_hd;
  DEV *dev_p;

  dev_p = mount_table + dev;
  desc_hd = dev_p->desc;
  /* Verification du disque ou on veut ecrire la table */
  if (dev_p->dev != dev) {
    fprintf(stderr,"Ecriture sur un disque inconnu: %d\n",dev);
    num_err = DISQUE_INCONNU;
    return -1;
  }
  /* Pose d'un verrou sur le bloc a ecrire */
  verrou = (struct flock*) malloc(sizeof(struct flock));
  verrou->l_type = F_WRLCK;
  verrou->l_whence = SEEK_SET;
  verrou->l_start = (off_t) TAILLE_BLOC;
  verrou->l_len = (off_t) TAILLE_BLOC;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Ecriture du bloc */
  if (lseek(desc_hd , (off_t) TAILLE_BLOC , SEEK_SET) == -1) {
    perror("lseek");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  dev_p->free_blocs_table[MAX_FREE_BLOCS] = dev_p->cur_p_bloc;
  if (write(desc_hd,dev_p->free_blocs_table,TAILLE_BLOC) != TAILLE_BLOC) {
    perror("write");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }
  /* Deverrouillage */
  verrou->l_type = F_UNLCK;
  if (fcntl(desc_hd,F_SETLKW,verrou) == -1) {
    perror("fcntl");
    free(verrou);
    num_err = ERREUR_INCONNUE; 
    return -1;
  }

  free(verrou);
  return 0;    
}

/* Recherche la reference 'reference' qui peut etre un chemin relatif
   ou absolu. 
   Retourne dans une structure n_noeud_s le numero de son noeud et de son peripherique
   si elle existe sinon met le champ num a -1.
   Genere directement des erreurs de type CHEMIN_INCORRECT si le chemin de la reference 
   passee en parametre est incorrect ou ENTREE_REPERTOIRE_INCONNUE si le chemin est 
   correct mais le fichier/rep designe par la reference n'existe pas. */
n_noeud_s find_ref (char* reference) {
  char* ref;
  char** tab_ref;
  int taille_tab = 1;
  char* s;
  n_noeud_s noeud;
  int i;
  int fait = 0;
  DESC *desc_p;

  /* Recopie de la reference, si elle est vide on la remplace par '.' */
  if (strlen(reference) == 0) {
    ref = ".";
  }
  else {
    ref = (char*) malloc(strlen(reference)+1);
    if (strcpy(ref,reference) == NULL) {
      noeud.num = -1;
      num_err = ERREUR_INCONNUE;
      free(ref);
      return noeud;
    }
  }
  /* On supprime '/' s'il est en dernier et que ref!="/" */
  if ((s=strrchr(ref,'/')) == ref+strlen(ref)-1 && strlen(ref) != 1) {
    *s = '\0';
  }
  /* On decompose ref */
  if (strcmp(ref,"/") != 0) {
    s = ref;
    while ((s=strchr(s,'/')) != NULL) {
      taille_tab ++;
      s++;
    }
  }
  tab_ref = (char**) malloc(sizeof(char*)*taille_tab);
  s = ref;
  for (i=0 ; i<taille_tab ; i++) {
    tab_ref[i] = s;
    while (*s != '/' && *s != '\0') {
      s ++;
    }
    *s = '\0';
    s ++;
  }
  /* Recherche du num de noeud de la ref */
  if (strlen(tab_ref[0]) == 0) { /* La ref est un chemin absolu */
    noeud.num = ROOT_NODE;
    noeud.dev = ROOT_DEV;
    fait ++;
  }
  else { /* La ref est un chemin relatif */
    desc_p = desc_table + desc_cur_dir;
    noeud.num = desc_p->node;
    noeud.dev = desc_p->dev;
  }
  while (fait != taille_tab) {
    if (find_name_in_node(tab_ref[fait],&noeud) == -1) {
      if (fait < taille_tab-1) /* Erreur survenue dans le chemin */
	num_err = CHEMIN_INCORRECT;
      else  /* erreur sur le dernier elt de la reference */
	num_err = ENTREE_REPERTOIRE_INCONNUE;
      noeud.num = -1;
      free(ref);
      free(tab_ref);
      return noeud;
    }
    fait ++;
  }

  free(tab_ref);
  free(ref);
  return noeud;
}

/* Recherche la ref 'name' dans le rep de num de noeud et de disque 'node'
   et remplace les champs de node par les num de noeud et de disque de la 
   ref si elle est trouvee.
   Retourne 0 en cas de succes, -1 sinon.
   Genere des erreurs de type ERREUR_INCONNUES, ENTREE_REPERTOIRE_INCONNUE
   si 'name' n'existe pas, TYPE_INCORRECT si le noeud 'node' ne
   correspond pas a un rep ou DISQUE_INCONNU si le disque du noeud n'existe pas */
int find_name_in_node(char* name, n_noeud_p node) {
  noeud_p np;
  n_bloc_s bloc;
  void* buf_bloc;
  void* indirect_bloc;
  short *short_p;
  entrep_p entree_p;
  int i,j,k,max,fin,trouve=0;
  short dev;
  n_noeud_s noeud_pere;

  dev = node->dev;
  noeud_pere.num = node->num;
  noeud_pere.dev = node->dev;
  /* Lecture du noeud 'node' */
  np = (noeud_p) malloc(sizeof(noeud_s));
  if (read_node(np, *node) == -1) {
    perror("read_node");
    free(np);
    return -1;
  }
  /* Le fichier correspondant a 'node' doit etre un rep */
  if (np->type != REP || np->etat != NOEUD_UTILISE) {
    num_err = TYPE_INCORRECT;
    free(np);
    return -1;
  }
  /* Traitement du cas evidant '.' */
  if (strcmp(name,".") == 0)
    return 0;
  /* Recherche de 'name' ds le rep */
  max = np->taille / TAILLE_BLOC;
  if (np->taille % TAILLE_BLOC != 0) {
    max ++;
  }
  buf_bloc = malloc(TAILLE_BLOC);
  indirect_bloc = malloc(TAILLE_BLOC);
  bloc.dev = dev;
  for (i=0 ; i<max ; i++) {
    if (i < 11) { /* Pas d'indirection pour les 11 1er blocs */
      bloc.num = np->blocs[i];
    }
    else { /* Au moins 1 niv d'indirection pour les blocs suivants */
      if ((i-11)%(TAILLE_BLOC/sizeof(short)) == 0) {
	if (i == 11)
	  bloc.num = np->blocs[i];
	else 
	  bloc.num = *short_p;
	/* Lecture du bloc d'indirection */
	if (read_bloc(indirect_bloc, bloc) == -1) {
	  perror("read_bloc");
	  free(np);
	  free(buf_bloc);
	  free(indirect_bloc);
	  return -1;
	}
	short_p = indirect_bloc;
      }
      bloc.num = *short_p;
      short_p ++;
    }
    /* Lecture du bloc contenant les entrees du rep */
    if (read_bloc(buf_bloc, bloc) == -1) {
      perror("read_bloc");
      free(np);
      free(buf_bloc);
      free(indirect_bloc);
      return -1;
    }
    /* Lecture des entrees du bloc */
    entree_p = buf_bloc;
    if (i == max-1) {
      fin = (np->taille / sizeof(entrep_s)) % ENTREES_PAR_BLOC;
      if (fin == 0)
	fin = ENTREES_PAR_BLOC;
    }
    else {
      fin = ENTREES_PAR_BLOC;
    }
    for (j=0 ; j<fin ; j++) {
      if ((node->num=entree_p->numero) != ENTREE_EFFACEE) {
	if (strcmp(entree_p->nom,name) == 0) {
	  /* On a trouve l'entree recherchee */
	  /* On verifie si on change de peripherique */
	  if (node->num == ROOT_NODE) {
	    /* Si 'name' est .., noeud_pere est un fils d'une racine de disque monte */
	    if (strcmp(name,"..") == 0) {
	      /* Si c'est lui-meme une racine de disque monte */
	      if (noeud_pere.num == ROOT_NODE) {
		node->dev = ((mount_table+noeud_pere.dev)->pere).dev;
	      }
	      /* Si c'est un repertoire ordinaire, le num de peripherique est deja le bon */
	      free(np);
	      free(buf_bloc);
	      free(indirect_bloc);
	      return 0;
	    }
	    /* Recherche du peripherique ayant pour reference 'name' */
	    for (k=0 ; k<MAX_MOUNT_DISK ; k++) {
	      if (mount_table[k].dev!=DEV_LIBRE) {
		if (strcmp(name,mount_table[k].ref)==0) {
		  /* On verifie que ce peripherique est bien monte sur noeud_pere */
		  if (mount_table[k].pere.num == noeud_pere.num &&
		      mount_table[k].pere.dev == noeud_pere.dev ) {
		    node->dev = mount_table[k].dev;
		    trouve = 1;
		    break;
		  }
		}
	      }
	    }
	    /* En a-t-on trouve un ? */
	    if (trouve != 1) { /* non */
	      num_err = DISQUE_INCONNU;
	      free(np);
	      free(buf_bloc);
	      free(indirect_bloc);
	      return -1;
	    }
	  }
	  free(np);
	  free(buf_bloc);
	  free(indirect_bloc);
	  return 0;
	}
      }
      entree_p ++;
    }
  }

  free(np);
  free(buf_bloc);
  free(indirect_bloc);
  num_err = ENTREE_REPERTOIRE_INCONNUE;
  return -1;
}

/* Supprime le noeud 'noeud' si plus aucune ref ne pointe sur lui 
   => il faut que 'liens' == 0 
   La suppression d'un noeud supprime automatiquement tous les 
   blocs de donnees associes a ce noeud et ajoute eventuellement 
   le numero de noeud a la table des noeuds libres.
   Retourne 0 en cas de succes, -1 sinon.
   Genere des erreurs de type RESTE_DES_LIENS si 'liens' != 0 
   ou ERREUR_INCONNUE */
int supp_node (n_noeud_s noeud) {
  n_bloc_s num_bloc_indirection, num_bloc;
  noeud_p np;
  int i,nb_bloc;
  void *indirect_bloc;
  short *short_p;
  short dev;
  DEV *dev_p;

  dev = noeud.dev;
  dev_p = mount_table + dev;  
  /* Lecture du noeud */
  np = (noeud_p) malloc(sizeof(noeud_s));
  if (read_node(np,noeud) == -1) {
    perror("read_node");
    free(np);
    return -1;
  }
  /* On verifie que le noeud n'a plus de lien */
  if (np->liens != 0) { 
    fprintf(stderr,"Impossible de supprimer le noeud %d,",noeud.num);
    fprintf(stderr,"il possede encore %d lien(s)\n",np->liens); 
    free(np); 
    num_err = RESTE_DES_LIENS; 
    return -1; 
  } 
  /* Suppression des blocs de donnees associes a ce noeud */ 
  nb_bloc = np->taille / TAILLE_BLOC;
  if (np->taille % TAILLE_BLOC != 0) {
    nb_bloc ++;
  }
  indirect_bloc = malloc(TAILLE_BLOC);
  num_bloc_indirection.dev = dev;
  num_bloc.dev = dev;
  /* Recherche des blocs a supprimer */
  for (i=0 ; i<nb_bloc ; i++) {
    if (i < 11) { /* Pas d'indirection pour les 11 1er blocs */
      num_bloc.num = np->blocs[i];
    }
    else { /* Au moins 1 niv d'indirection pour les blocs suivants */
      if ((i-11)%(TAILLE_BLOC/sizeof(short)) == 0) {
	if (i == 11)
	  num_bloc_indirection.num = np->blocs[i];
	else 
	  num_bloc_indirection.num = *short_p;
	/* Lecture du bloc d'indirection */
	if (read_bloc(indirect_bloc, num_bloc_indirection) == -1) {
	  perror("read_bloc");
	  free(np);
	  free(indirect_bloc);
	  return NULL;
	}
	short_p = indirect_bloc;
      }
      num_bloc.num = *short_p;
      short_p ++;
    }
    if (supp_bloc(num_bloc) == -1) { 
      perror("supp_bloc"); 
      fprintf(stderr,"%ieme bloc (%i) non libere\n",i,num_bloc.num); 
    } 
  }
  free(indirect_bloc);
  np->taille = 0; 
  /* On libere le noeud */ 
  np->etat = NOEUD_LIBRE; 
  if (write_node(np,noeud) == -1) {
    free(np); 
    perror("write_node"); 
    return -1; 
  } 
  /* Mise a jour de la table des noeuds libres si necessaire */ 
  if (noeud.num < dev_p->free_nodes_table[MAX_FREE_NODES-1]) { 
    if (dev_p->cur_p_node == 0) { 
      for (i=MAX_FREE_NODES ; i>0 ; i--) { 
	dev_p->free_nodes_table[i] = dev_p->free_nodes_table[i-1]; 
      } 
    } 
    else 
      dev_p->cur_p_node --; 
    dev_p->free_nodes_table[dev_p->cur_p_node] = noeud.num; 
    if (write_free_nodes_table(dev) == -1) {
      perror("write_free_nodes_table");
      free(np); 
      fprintf(stderr,"Impossible de modifier la table des noeuds libres"); 
      return -1; 
    } 
  } 
  
  free(np); 
  return 0; 
} 

/* Retourne le numero d'un noeud libre sur le disque 'dev' et supprime celui-ci
   de la liste des noeuds libres en incrementant 'cur_p_noeud'. Si celui-ci sort  
   de la liste, on la reremplit en partant de la fin pour conserver son
   ordre (il n'y a pas de numero de noeud libre plus petit que ceux qui 
   se trouvent dans la table des noeuds libres).
   Met le champ 'num' de la struture retournee a -1 en cas d'erreur.
   Genere des erreurs de type ERREUR_INCONNUE ou PLUS_DE_NOEUD_LIBRE
   si tous les noeuds sont occupes */ 
n_noeud_s new_node (short dev) { 
  n_noeud_s num_noeud; 
  n_noeud_s num_noeud_libre; 
  int i,j; 
  noeud_p np;
  DEV *dev_p;

  dev_p = mount_table + dev;
  num_noeud.dev = dev;
  /* Verification du disque ou on veut ajouter un noeud */
  if (dev_p->dev != dev) {
    fprintf(stderr,"Acces a un disque inconnu: %d\n",dev);
    num_err = DISQUE_INCONNU;
    num_noeud.num = -1;
    return num_noeud; 
  }
  /* On verifie qu'il reste encore des noeuds libres */ 
  if (dev_p->cur_p_node == MAX_FREE_NODES) { 
    num_err = PLUS_DE_NOEUD_LIBRE; 
    fprintf(stderr,"Plus de noeud libre"); 
    num_noeud.num = -1;
    return num_noeud; 
  } 
  num_noeud.num = dev_p->free_nodes_table[dev_p->cur_p_node]; 
  dev_p->cur_p_node ++; 
  /* Remplissage de la table des noeuds libres si necessaire */ 
  np = (noeud_p) malloc(sizeof(noeud_s)); 
  num_noeud_libre.dev = dev;
  if (dev_p->cur_p_node == MAX_FREE_NODES) { 
    /* On recherche de nouveaux noeuds libres a partir de num_noeud+1 */ 
    for (i=num_noeud.num+1; i<NOEUDS_PAR_BLOC*TAILLE_ZONE_NOEUDS && dev_p->cur_p_node!=0; i++) { 
      num_noeud_libre.num = i;
      if (read_node(np,num_noeud_libre) == -1) { 
	perror("read_node"); 
	free(np); 
	num_noeud.num = -1;
	return num_noeud; 
      } 
      /* Si on en trouve un, on le rajoute a la fin de la table 
	 en decalant tous les autres vers le bas */ 
      if (np->etat == NOEUD_LIBRE) { 
	dev_p->cur_p_node --; 
	for (j=dev_p->cur_p_node ; j<MAX_FREE_NODES-1 ; j++) { 
	  dev_p->free_nodes_table[j] = dev_p->free_nodes_table[j+1]; 
	} 
	dev_p->free_nodes_table[MAX_FREE_NODES-1] = i; 
      } 
    } 
  } 
  free(np); 
  /* Sauvegarde des modifications */ 
  if (write_free_nodes_table(dev) == -1) { 
    perror("write_free_nodes_table"); 
    num_noeud.num = -1;
    return num_noeud; 
  } 
 
  return num_noeud; 
} 
 
/* Retourne un numero de bloc libre sur le disque 'dev' et supprime celui-ci
   de la liste des blocs libres en incrementant 'cur_p_bloc'. Si celui-ci sort  
   de la liste contenue dans le 2eme bloc, on reremplit celui-ci par  
   le bloc suivant de la liste des blocs libres.
   Met le champ 'num' de la struture retournee a -1 en cas d'erreur.
   Genere des erreurs de type ERREUR_INCONNUE ou PLUS_DE_BLOC_LIBRE
   si tous les blocs sont occupes  */ 
n_bloc_s new_bloc (short dev) { 
  n_bloc_s num_bloc; 
  short *short_tab; 
  void * buf_bloc; 
  int i; 
  DEV *dev_p;

  dev_p = mount_table + dev;
  num_bloc.dev = dev;
  /* Verification du disque ou on veut ajouter un bloc */
  if (dev_p->dev != dev) {
    fprintf(stderr,"Acces a un disque inconnu: %d\n",dev);
    num_err = DISQUE_INCONNU;
    num_bloc.num = -1;
    return num_bloc; 
  }
  /* On verifie qu'il reste encore des blocs libres */ 
  if (dev_p->cur_p_bloc == MAX_FREE_BLOCS) { 
    num_err = PLUS_DE_BLOC_LIBRE; 
    fprintf(stderr,"Plus de bloc libre"); 
    num_bloc.num = -1;
    return num_bloc; 
  } 
  num_bloc.num = dev_p->free_blocs_table[dev_p->cur_p_bloc]; 
  dev_p->cur_p_bloc ++; 
  /* Remplissage de la table des blocs libres si necessaire */ 
  buf_bloc = malloc(TAILLE_BLOC); 
  if (dev_p->cur_p_bloc == MAX_FREE_BLOCS) { 
    if (read_bloc(buf_bloc,num_bloc) == -1) { 
      perror("read_bloc");
      free(buf_bloc); 
      num_bloc.num = -1;
      return num_bloc; 
    } 
    short_tab = buf_bloc; 
    for (i=0 ; i<MAX_FREE_BLOCS ; i++) { 
      dev_p->free_blocs_table[i] = short_tab[i]; 
    } 
    dev_p->cur_p_bloc = 0;
  } 
  /* Sauvegarde sur le disque des changements */ 
  if (write_free_blocs_table(dev) == -1) {
    perror(" write_free_blocs_table");
    free(buf_bloc); 
    num_bloc.num = -1;
    return num_bloc; 
  } 
  
  free(buf_bloc); 
  return num_bloc; 
} 
 
/* Supprime le bloc 'bloc'. Ne verifie pas si il est toujours utilise
   par un noeud. Une fois supprime, son numero est ajoute a la liste des blocs libres.
   Si la table des blocs libre (contenue dans le 2eme bloc du disque) est pleine,
   elle est recopiee dans ce bloc et on y met que le numero de ce bloc.
   Retourne 0 en cas de succes, -1 sinon.
   Ne genere que des erreurs de type ERREUR_INCONNUE. */ 
int supp_bloc (n_bloc_s bloc) { 
  int i; 
  short *short_p; 
  void *buf_bloc; 
  DEV *dev_p;
  short dev;

  dev = bloc.dev;
  dev_p = mount_table + dev;
  /* Verification du disque ou on veut supp un bloc */
  if (dev_p->dev != bloc.dev) {
    fprintf(stderr,"Acces a un disque inconnu: %d\n",bloc.dev);
    num_err = DISQUE_INCONNU;
    return -1;
  }
  /* Verification du bloc qu'on veut ecrire */
  if (bloc.num < 0  || bloc.num >= TAILLE_ZONE_DONNEES) {
    fprintf(stderr,"Suppressoion d'un bloc inexistant: %d\n",bloc.num);
    num_err = MAUVAISE_POSITION;
    return -1;
  }
  buf_bloc = calloc (TAILLE_BLOC,1); 
  /* Si la table des blocs libres est pleine on la recopie dans 
     le bloc 'bloc' et on y met 'bloc' en dernier et on fait pointer 
     cur_p_bloc dessus */ 
  if (dev_p->cur_p_bloc == 0) { 
    short_p = buf_bloc; 
    for (i=0 ; i<MAX_FREE_BLOCS ; i++) { 
      *short_p = dev_p->free_blocs_table[i]; 
      short_p ++; 
    } 
    *short_p = 0; 
    dev_p->cur_p_bloc = MAX_FREE_BLOCS; 
  } 
  if (write_bloc(buf_bloc,bloc) == -1) { 
    free(buf_bloc); 
    perror("write_bloc"); 
    return -1; 
  } 
  free(buf_bloc); 
  dev_p->cur_p_bloc --; 
  dev_p->free_blocs_table[dev_p->cur_p_bloc] = bloc.num;
  /* Sauvegarde des modifications */ 
  if (write_free_blocs_table(dev) == -1) { 
    perror("write_free_blocs_table"); 
    return -1; 
  } 
  
  return 0; 
} 
 
/* Ajoute une entree au rep dont le noeud est 'num_noeud' retourne un pointeur sur ce
   noeud modifie ou NULL en cas d'erreur.
   Genere directement des erreurs de type ERREUR_INCONNUE, ENTREE_EXISTANTE 
   si une entree ayant le meme nom existe deja ds le rep ou TYPE_INCORRECT si le
   numero de noeud 'num_noeud' n'est pas celui d'un rep. Genere aussi
   PLUS_DE_BLOC_LIBRE a partir de new_bloc */
noeud_p new_entry (n_noeud_s num_noeud,entrep_p entree) { 
  void* buf_bloc;
  void* buf_bloc_find = NULL;
  void* indirect_bloc;
  noeud_p np; 
  entrep_p entree_p; 
  n_bloc_s num_bloc,num_bloc_find,num_new_bloc,num_bloc_indirection,num_new_bloc_indirection; 
  short *short_p;
  int i,j,nb_bloc,max;
  short dev;

  dev = num_noeud.dev;
  np = (noeud_p) malloc(sizeof(noeud_s)); 
  /* Lecture du noeud a modifier */ 
  if (read_node(np,num_noeud) == -1) { 
    perror("read_node"); 
    free(np); 
    return NULL; 
  } 
  /* Verification qu'il s'agit bien d'un rep */ 
  if (np->type != REP) { 
    fprintf(stderr,"Le noeud %d ne correspond pas a un rep",num_noeud.num); 
    free(np); 
    num_err = TYPE_INCORRECT; 
    return NULL; 
  } 
  /* On cherche une entree du repertoire qui serait vide et 
     on verifie qu'il n'y a pas deja une entree ayant ce nom */
  nb_bloc = np->taille / TAILLE_BLOC;
  if (np->taille % TAILLE_BLOC != 0) {
    nb_bloc ++;
  }
  buf_bloc = malloc(TAILLE_BLOC);
  indirect_bloc = malloc(TAILLE_BLOC);
  num_bloc_indirection.dev = dev;
  num_bloc.dev = dev;
  for (i=0 ; i<nb_bloc ; i++) {
    if (i < 11) { /* Pas d'indirection pour les 11 1er blocs */
      num_bloc.num = np->blocs[i];
    }
    else { /* Au moins 1 niv d'indirection pour les blocs suivants */
      if ((i-11)%(TAILLE_BLOC/sizeof(short)) == 0) {
	if (i == 11)
	  num_bloc_indirection.num = np->blocs[i];
	else 
	  num_bloc_indirection.num = *short_p;
	/* Lecture du bloc d'indirection */
	if (read_bloc(indirect_bloc, num_bloc_indirection) == -1) {
	  perror("read_bloc");
	  free(np);
	  free(buf_bloc);
	  free(indirect_bloc);
	  return NULL;
	}
	short_p = indirect_bloc;
      }
      num_bloc.num = *short_p;
      short_p ++;
    }
    /* Lecture du bloc contenant les entrees du rep */
    if (read_bloc(buf_bloc, num_bloc) == -1) {
      perror("read_bloc");
      free(np);
      free(buf_bloc);
      free(indirect_bloc);
      return NULL;
    }
    /* Lecture des entrees du bloc */
    entree_p = buf_bloc;
    if (i == nb_bloc-1) {
      max = (np->taille / sizeof(entrep_s)) % ENTREES_PAR_BLOC;
      if (max == 0)
	max = ENTREES_PAR_BLOC;
    }
    else {
      max = ENTREES_PAR_BLOC;
    }
    for (j=0 ; j<max ; j++) {
      if ((entree_p->numero) == ENTREE_EFFACEE && buf_bloc_find == NULL) {
	/* On a trouve une entree vide */
	/* On memorise ou se trouve cette entree */
	*entree_p = *entree;
	buf_bloc_find = malloc(TAILLE_BLOC);
	memcpy(buf_bloc_find, buf_bloc, (size_t)TAILLE_BLOC);
	num_bloc_find.dev = num_bloc.dev;
	num_bloc_find.num = num_bloc.num;
	/* On continue a balayer toutes les entrees du rep pour
	   verifier qu'il n'y a pas deja une entree ayant le meme nom */
      }
      else {
	if (strcmp(entree_p->nom,entree->nom) == 0) {
	  /* Une entree portant le meme nom existe deja */
	  free(np);
	  free(buf_bloc);
	  if (buf_bloc_find != NULL)
	    free(buf_bloc_find);
	  free(indirect_bloc);
	  num_err = ENTREE_EXISTANTE;
	  return NULL;
	}
      }
      entree_p ++;
    }
  }
  if (buf_bloc_find != NULL) { /* On a trouve une entree vide */
    if (write_bloc(buf_bloc_find, num_bloc_find) == -1) {
      perror("write_bloc");
      free(np);
      free(buf_bloc);
      free(buf_bloc_find);
      free(indirect_bloc);
      return NULL;
    }
    free(buf_bloc);
    free(buf_bloc_find);
    free(indirect_bloc);
    return np;
  }
  /* Aucune des entrees du rep n'est vide : On rajoute l'entree a la fin */
  np->taille += sizeof(entrep_s);
  /* On regarde si le dernier bloc de donnees est plein */ 
  if (np->taille % TAILLE_BLOC == sizeof(entrep_s)) { /* Il est plein */ 
    /* Cree un nouveau bloc de donnees pour cette entree */
    entree_p = buf_bloc;
    *entree_p = *entree;
    num_new_bloc=new_bloc(dev);
    if (num_new_bloc.num == -1) {
      perror("new_bloc"); 
      free(buf_bloc); 
      free(indirect_bloc);
      free(np); 
      np->taille -= sizeof(entrep_s);
      return NULL; 
    } 
    if (write_bloc(buf_bloc,num_new_bloc) == -1) { 
      perror("write_bloc"); 
      free(buf_bloc); 
      free(indirect_bloc);
      free(np); 
      np->taille -= sizeof(entrep_s);
      return NULL; 
    }
    /* Rajoute le bloc a la liste des blocs du rep. On regarde si cela engendre 
       ou non un niveau supplementaire d'indirection */
    nb_bloc ++;
    if (nb_bloc < 12) { /* Pas d'indrection */
      np->blocs[nb_bloc-1] = num_new_bloc.num;
    }
    else { /* Il y a deja au moins 1 niveau d'indirection ou on va creer le 1er */
      if ((nb_bloc-12)%(TAILLE_BLOC/sizeof(short)-1) == 0) {
	/* Le dernier bloc d'indirection est plein ou nb_bloc==12
	   -> il faut creer un nouveau bloc d'indirection */
	/* Ajoute le nouveau bloc de donnees ds le nouveau bloc d'indirection */
	num_new_bloc_indirection = new_bloc(dev);
	if (num_new_bloc_indirection.num == -1) {
	  perror("new_bloc"); 
	  free(buf_bloc); 
	  free(indirect_bloc);
	  free(np); 
	  np->taille -= sizeof(entrep_s);
	  return NULL; 
	} 
	short_p = buf_bloc;
	*short_p = num_new_bloc.num;
	if (write_bloc(buf_bloc,num_new_bloc_indirection) == -1) { 
	  perror("write_bloc"); 
	  free(buf_bloc); 
	  free(indirect_bloc);
	  free(np); 
	  np->taille -= sizeof(entrep_s);
	  return NULL; 
	}
	/* Ajoute le nouveau bloc d'indirection ds le bloc d'indirection precedent */
	if (nb_bloc == 12) {
	  np->blocs[nb_bloc-1] = num_new_bloc_indirection.num;
	}
	else {
	  short_p = indirect_bloc;
	  short_p += TAILLE_BLOC / sizeof(short) -2;
	  *short_p = num_new_bloc_indirection.num;
	  if (write_bloc(indirect_bloc,num_bloc_indirection) == -1) { 
	    perror("write_bloc"); 
	    free(buf_bloc); 
	    free(indirect_bloc);
	    free(np); 
	    np->taille -= sizeof(entrep_s);
	    return NULL; 
	  }
	}
      }
      else { /* Le dernier bloc d'indirection n'est pas plein */
	short_p = indirect_bloc;
	short_p += (nb_bloc-12) % (TAILLE_BLOC/sizeof(short)-1);
	*short_p = num_new_bloc.num;
	if (write_bloc(indirect_bloc,num_bloc_indirection) == -1) { 
	  perror("write_bloc"); 
	  free(buf_bloc); 
	  free(indirect_bloc);
	  free(np); 
	  np->taille -= sizeof(entrep_s);
	  return NULL; 
	}
      }
    }
  } 
  else { /* Le dernier bloc n'est pas plein */ 
    /* Ajoute la nouvelle entree ds ce bloc */
    *entree_p = *entree; 
    if (write_bloc(buf_bloc,num_bloc) == -1) { 
      perror("write_bloc"); 
      free(buf_bloc); 
      free(indirect_bloc);
      free(np); 
      np->taille -= sizeof(entrep_s);
      return NULL; 
    } 
  }

  free(indirect_bloc);
  free(buf_bloc);
  if (write_node(np,num_noeud) == -1) {
    perror("write_node");
    free(np);
    return NULL;
  }
  return np;
}

/* Supprime une entree du rep de noeud 'num_noeud'. S'il ne s'agit pas de la derniere entree
   on la ermplace par la derniere entree. Apres cela ou directement s'il s'agit de la derniere
   entree, on supprime la derniere entree (en diminuant la taille du rep) et
   on supprime tous les blocs devenus inutils (bloc contenant l'entree et bloc 
   d'indirection) si il y en a.
   Retourne NULL en cas d'echec ou un pointeur sur le noeud modifie.
   Genere des erreurs de type ERREUR_INCONNUES, ENTREE_REPERTOIRE_INCONNUE
   si 'entree' n'existe pas ds le rep ou TYPE_INCORRECT si le noeud 
   'num_noeud' ne correspond pas a un rep */
noeud_p supp_entry (n_noeud_s num_noeud,entrep_p entree) { 
  void *buf_bloc,*indirect_bloc,*buf_bloc_trouve;
  noeud_p np; 
  entrep_p entree_p,entree_trouvee_p; 
  n_bloc_s num_bloc,num_bloc_indirection,num_bloc_trouve; 
  short pos_entree_trouvee;
  short *short_p;
  int i,j,nb_bloc,max;
  short dev;

  dev = num_noeud.dev;
  np = (noeud_p) malloc(sizeof(noeud_s)); 
  buf_bloc_trouve = NULL;
  /* Lecture du noeud a modifier */ 
  if (read_node(np,num_noeud) == -1) { 
    perror("read_node"); 
    free(np); 
    return NULL; 
  } 
  /* Verification qu'il s'agit bien d'un rep */ 
  if (np->type != REP) { 
    fprintf(stderr,"Le noeud %d ne correspond pas a un rep",num_noeud.num); 
    free(np); 
    num_err = TYPE_INCORRECT; 
    return NULL; 
  } 
  /* On cherche l'entree du repertoire a supprimer */
  num_bloc_indirection.dev = dev;
  num_bloc.dev = dev;
  nb_bloc = np->taille / TAILLE_BLOC;
  if (np->taille % TAILLE_BLOC != 0) {
    nb_bloc ++;
  }
  buf_bloc = malloc(TAILLE_BLOC);
  indirect_bloc = malloc(TAILLE_BLOC);
  for (i=0 ; i<nb_bloc ; i++) {
    if (i < 11) { /* Pas d'indirection pour les 11 1er blocs */
      num_bloc.num = np->blocs[i];
    }
    else { /* Au moins 1 niv d'indirection pour les blocs suivants */
      if ((i-11)%(TAILLE_BLOC/sizeof(short)) == 0) {
	if (i == 11)
	  num_bloc_indirection.num = np->blocs[i];
	else 
	  num_bloc_indirection.num = *short_p;
	/* Lecture du bloc d'indirection */
	if (read_bloc(indirect_bloc, num_bloc_indirection) == -1) {
	  perror("read_bloc");
	  free(np);
	  free(buf_bloc);
	  free(indirect_bloc);
	  return NULL;
	}
	short_p = indirect_bloc;
      }
      num_bloc.num = *short_p;
      short_p ++;
    }
    /* Lecture du bloc contenant les entrees du rep */
    if (read_bloc(buf_bloc, num_bloc) == -1) {
      perror("read_bloc");
      free(np);
      free(buf_bloc);
      free(indirect_bloc);
      return NULL;
    }
    /* Lecture des entrees du bloc */
    entree_p = buf_bloc;
    if (i == nb_bloc-1) {
      max = (np->taille / sizeof(entrep_s)) % ENTREES_PAR_BLOC;
      if (max == 0)
	max = ENTREES_PAR_BLOC;
    }
    else {
      max = ENTREES_PAR_BLOC;
    }
    for (j=0 ; j<max ; j++) {
      if (entree_p->numero == entree->numero) { 
	if (strcmp(entree_p->nom,entree->nom) == 0) {
	  /* On a trouve l'entree a supprimer */
	  /* On regarde si c'est la derniere entree du rep */
	  if (i == nb_bloc-1 && j == max-1) {
	    /* On supprime l'entree */
	    np->taille -= sizeof(entrep_s);
	    /* On regarde si ca ne libere pas un bloc de donnees */
	    if ((np->taille/sizeof(entrep_s))%ENTREES_PAR_BLOC == 0) {
	      nb_bloc --;
	      if (supp_bloc(num_bloc) == -1) {
		perror("supp_bloc");
		free(np);
		free(buf_bloc);
		free(indirect_bloc);
		return NULL;
	      }
	      /* On regarde si ca ne libere pas un bloc d'indirection */
	      if ((nb_bloc-11)%(TAILLE_BLOC/sizeof(short)-1) == 0) {
		if (supp_bloc(num_bloc_indirection) == -1) {
		  perror("supp_bloc");
		  free(np);
		  free(buf_bloc);
		  free(indirect_bloc);
		  return NULL;
		}
	      }
	    }
	    /* Enregistrement des modifications du noeud */ 
	    if (write_node(np,num_noeud) == -1) { 
	      perror("write_node"); 
	      free(np); 
	      free(buf_bloc);
	      free(indirect_bloc);
	      return NULL; 
	    }
	    free(buf_bloc);
	    free(indirect_bloc);
	    return np;
	  }
	  else { /* L'entree supprimee n'est pas la derniere */
	    /* On memorise sa position et son bloc pour la remplasser par la derniere entree */
	    pos_entree_trouvee = j;
	    buf_bloc_trouve = malloc(TAILLE_BLOC);
	    memcpy(buf_bloc_trouve, buf_bloc, (size_t)TAILLE_BLOC);
	    num_bloc_trouve = num_bloc;
	  }
	}
      }
      entree_p ++;
    } /* fin du for(j=0 ; j<max ; j++) */
  } /* fin du for(i=0 ; i<nb_bloc ; i++) */
  if(buf_bloc_trouve != NULL){/* L'entree supprimee a ete trouvee mais n'est pas la derniere */
    /* Recopie de la derniere entree a la place de l'entree supprimee */
    entree_p --;
    entree_trouvee_p = (entrep_p)buf_bloc_trouve + pos_entree_trouvee;
    *entree_trouvee_p = *entree_p;
    if (write_bloc(buf_bloc_trouve,num_bloc_trouve) == -1) {
      free(buf_bloc_trouve);
      free(np); 
      free(buf_bloc);
      free(indirect_bloc);
      return NULL;
    }
    free(buf_bloc_trouve);
    /* Supression de la derniere entree */		
    np->taille -= sizeof(entrep_s);
    /* On regarde si ca ne libere pas un bloc de donnees */
    if ((np->taille/sizeof(entrep_s))%ENTREES_PAR_BLOC == 0) {
      nb_bloc --;
      if (supp_bloc(num_bloc) == -1) {
	perror("supp_bloc");
	free(np);
	free(buf_bloc);
	free(indirect_bloc);
	return NULL;
      }
      /* On regarde si ca ne libere pas un bloc d'indirection */
      if ((nb_bloc-11)%(TAILLE_BLOC/sizeof(short)-1) == 0) {
	if (supp_bloc(num_bloc_indirection) == -1) {
	  perror("supp_bloc");
	  free(np);
	  free(buf_bloc);
	  free(indirect_bloc);
	  return NULL;
	}
      }
    }
    /* Enregistrement des modifications du noeud */ 
    if (write_node(np,num_noeud) == -1) { 
      perror("write_node"); 
      free(np); 
      free(buf_bloc);
      free(indirect_bloc);
      return NULL; 
    }
    free(buf_bloc);
    free(indirect_bloc);
    return np;
  }
  free(np); 
  free(buf_bloc);
  free(indirect_bloc);
  num_err = ENTREE_REPERTOIRE_INCONNUE;
  return NULL;
}

/* Recherche le num de bloc du (bloc_i)ieme bloc du noeud 'num_noeud'.
   Retourne dans le champ num: -1 en cas d'echec, le num de bloc sinon. 
   Genere des erreurs de type ERREUR_INCONNUE ou MAUVAISE_POSITION si
   'bloc_i' est superieur au nb de bloc du noeud 'num_noeud' */
n_bloc_s num_of_bloc_i (short bloc_i, n_noeud_s num_noeud) {
  int nb_bloc;
  short *short_tab;
  void* bloc_indirect;
  noeud_p np;
  n_bloc_s num_bloc;
  short dev;

  dev = num_noeud.dev;
  num_bloc.dev = dev;
  np = (noeud_p) malloc(sizeof(noeud_s)); 
  /* Lecture du noeud a modifier */ 
  if (read_node(np,num_noeud) == -1) { 
    perror("read_node"); 
    free(np);
    num_bloc.num = -1;
    return num_bloc; 
  } 
  /* Recherche le num de bloc */
  nb_bloc = np->taille / TAILLE_BLOC;
  if (np->taille % TAILLE_BLOC != 0) {
    nb_bloc ++;
  }
  if (bloc_i > nb_bloc || bloc_i < 0) {
    num_err = MAUVAISE_POSITION;
    free(np);
    num_bloc.num = -1;
    return num_bloc;
  }
  if (bloc_i < 11) {
    num_bloc.num = np->blocs[bloc_i];
    return num_bloc;
  }
  else {
    bloc_indirect = malloc(TAILLE_BLOC);
    bloc_i -= 11;
    num_bloc.num = np->blocs[11];
    if (read_bloc(bloc_indirect,num_bloc) == -1) {
      free(bloc_indirect);
      free(np);
      num_bloc.num = -1;
      return num_bloc;
    }
    short_tab = bloc_indirect;
    while (bloc_i > (TAILLE_BLOC/sizeof(short)-1)) {
      num_bloc.num = short_tab[TAILLE_BLOC/sizeof(short)-1];
      if (read_bloc(bloc_indirect,num_bloc) == -1) {
	free(bloc_indirect);
	free(np);
	num_bloc.num = -1;
	return num_bloc;
      }
      bloc_i -= (TAILLE_BLOC/sizeof(short)-1);
    }
    num_bloc.num = short_tab[bloc_i];

    free(np);
    free(bloc_indirect);
    return num_bloc;
  }
}

/* Vide le buffer associe au descripteur 'desc' (en le sauvegardant
   si le fichier n'a pas ete ouvert en lecture seule) et le remplit
   avec le (pos_new_bloc)ieme bloc du fichier.
   Retourne 0 en cas de succes, -1 sinon.
   Ne genere directement qu'une erreur de type DESCRIPTEUR_INCONNU si le
   descripteur passe en parametre est vide. Les autres erreurs proviennent
   de num_of_bloc_i, write_bloc et read_bloc */
int change_buf (desc_t desc,short pos_new_bloc) {
  n_bloc_s num_bloc;
  n_noeud_s num_noeud;
  DESC* desc_p;

  desc_p = &desc_table[desc];
  num_noeud.dev = desc_p->dev;
  num_noeud.num = desc_p->node;
  /* Verifie que le descripteur existe bien */
  if (desc_p->desc == DESC_LIBRE) {
    num_err = DESCRIPTEUR_INCONNU;
    return -1;
  }
  /* On sauvegarde automatiquement le buffer si le mode d'ouverture
     n'est pas en lecture seule */
  if (((desc_p->mode)&M_ECR) == M_ECR) {
    /* Si le fichier est vide, il n'y a rien a enregister */
    if (desc_p->stat->taille != 0) {
      num_bloc = num_of_bloc_i(desc_p->cur_bloc,num_noeud);
      if (num_bloc.num == -1) {
	return -1;
      }
      if (write_bloc(desc_p->buffer,num_bloc) == -1) {
	return -1;
      }
    }
  }
  /* Chargement du nouveau bloc */
  num_bloc = num_of_bloc_i(pos_new_bloc, num_noeud);
  if (num_bloc.num == -1) {
    return -1;
  }
  if (read_bloc(desc_p->buffer,num_bloc) == -1) {
    return -1;
  }

  return 0;
}

/* Ajoute le bloc de numero 'num_added_bloc' a la fin de la liste des blocs du noeud
   'num_noeud'. On precise en parametre la taille du noeud avant modification.
   Retourne -1 en cas d'echec, 0 sinon.
   Ne genere que des erreurs de type ERREUR_INCONNUE */
int add_bloc_in_node (n_noeud_s num_noeud, short num_added_bloc, short old_size) {
  int nb_bloc,i=0;
  short *short_tab;
  void* bloc_indirect;
  void* new_bloc_indirect;
  noeud_p np;
  n_bloc_s num_bloc,num_bloc_indirection;
  short dev;

  dev = num_noeud.dev;
  num_bloc.dev = dev;
  np = (noeud_p) malloc(sizeof(noeud_s)); 
  /* Lecture du noeud a modifier */ 
  if (read_node(np,num_noeud) == -1) { 
    perror("read_node"); 
    free(np);
    return -1; 
  } 
  /* Recherche le num de bloc */
  nb_bloc = old_size / TAILLE_BLOC;
  if (old_size % TAILLE_BLOC != 0) {
    nb_bloc ++;
  }
  if (nb_bloc < 11) {
    np->blocs[nb_bloc] = num_added_bloc;
    if (write_node(np,num_noeud) == -1) {
      free(bloc_indirect);
      free(np);
      return -1;
    }
    free(np);
    return 0;
  }
  else {
    bloc_indirect = malloc(TAILLE_BLOC);
    nb_bloc -= 11;
    short_tab = bloc_indirect;
    num_bloc.num = np->blocs[11];
    for (i=0 ; nb_bloc >= (TAILLE_BLOC/sizeof(short)-1) ; i++) {
      if (i != 0) {
	num_bloc.num = short_tab[TAILLE_BLOC/sizeof(short)-1];
      }
      if (read_bloc(bloc_indirect,num_bloc) == -1) {
	free(bloc_indirect);
	free(np);
	return -1;
      }
      nb_bloc -= (TAILLE_BLOC/sizeof(short)-1);
    }
    /* Ajout d'un bloc d'indirection si necessaire */
    if (nb_bloc == 0) {
      /* Creation d'un niv d'indirection */
      num_bloc_indirection = new_bloc(dev);
      if (num_bloc_indirection.num == -1) {
	free(bloc_indirect);
	free(np);
	return -1;
      }
      if (i == 0) { /* 1er niv d'indirection */
	np->blocs[11] = num_bloc_indirection.num;
	if (write_node(np,num_noeud) == -1) {
	  free(bloc_indirect);
	  free(np);
	  return -1;
	}
      }
      else { /* Le fichier avait deja au moins 1 niv d'indirection */
	short_tab[TAILLE_BLOC/sizeof(short)-1] = num_bloc_indirection.num;
	if (write_bloc(bloc_indirect,num_bloc) == -1) {
	  free(bloc_indirect);
	  free(np);
	  return -1;
	}
      }
      new_bloc_indirect = malloc(TAILLE_BLOC);
      *((short*)new_bloc_indirect) = num_added_bloc;
      if (write_bloc(new_bloc_indirect,num_bloc_indirection) == -1) {
	free(bloc_indirect);
	free(np);
	free(new_bloc_indirect);
	return -1;
      }
      free(new_bloc_indirect);
    }
    else { /* Le dernier bloc d'indirection n'est pas plein */
      if (i != 0) { /* Le fichier a au moins 2 niv d'indirection */
	num_bloc.num = short_tab[TAILLE_BLOC/sizeof(short)-1];
      }
      if (read_bloc(bloc_indirect,num_bloc) == -1) {
	free(bloc_indirect);
	free(np);
	return -1;
      }
      short_tab[nb_bloc] = num_added_bloc;
      if (write_bloc(bloc_indirect,num_bloc) == -1) {
	free(bloc_indirect);
	free(np);
	return -1;
      }
    }

    free(np);
    free(bloc_indirect);
    return 0;
  } /* else */
}
