#ifndef LIB_MINIFS
#define LIB_MINIFS

#include <stdlib.h>
#include <sys/types.h>

/**********************
  Definition des types
***********************/

typedef short desc_t;
typedef int pos_t;

/**********************************
  Constantes et structure imposees
***********************************/

/* Caracteristiques du disque */
#define TAILLE_BLOC 128           /* en octets */
#define TAILLE_ZONE_DESCRIPTION 2 /* en blocs */
#define TAILLE_ZONE_NOEUDS 12     /* en blocs */
#define TAILLE_ZONE_DONNEES 240   /* en blocs */
#define NOM_DISQUE_RACINE "disk1" /* disque racine */

/* Structure des noeuds stockes ds la zone des noeuds */
struct noeud {
  short etat;
  short type;
  short taille;
  short liens;
  short blocs[12];
};
typedef struct noeud noeud_s;
typedef struct noeud *noeud_p;

/* Constante et macros de la zone des noeuds */
#define NUMERO_BLOC_DEBUT_ZONE_NOEUDS 2
#define NOEUDS_PAR_BLOC (TAILLE_BLOC/sizeof(struct noeud))
#define NOEUD2BLOC(num_noeud) (num_noeud/NOEUDS_PAR_BLOC+NUMERO_BLOC_DEBUT_ZONE_NOEUDS)
#define NOEUDINBLOC(num_noeud) (num_noeud%NOEUDS_PAR_BLOC)
 
/* Structure de description des fichiers */
struct status{
  short etat;
  short type;
  short taille;
  short liens;
};
typedef struct status *stat_p;

/* Constantes et macros assossiees aux champs de stat_p */
#define NOEUD_LIBRE 0
#define NOEUD_UTILISE 1
#define EST_LIBRE(pn) ((pn)->etat==NOEUD_LIBRE)
#define FIC 0
#define REP 1
#define IS_REP(pn) ((pn)->type==REP)

/* Structure d'entrees des repertoires */
#define ENTREE_EFFACEE -1
struct entree_repertoire {
  char nom[30];
  short numero;
};
typedef struct entree_repertoire entrep_s;
typedef struct entree_repertoire *entrep_p;
#define ENTREES_PAR_BLOC (TAILLE_BLOC/sizeof(entrep_s))

/* Mode d'ouverture des fichiers */
#define M_LEC 1
#define M_ECR 2
#define M_EL 3
#define M_CRE 4
#define M_ZER 8

/* Point de depart possible dans lseek */
#define DEBUT 1
#define FIN 2
#define COURANT 4

/*******************************************************
  Constantes, structures et variables globales ajoutees
********************************************************/

/* Description des erreurs */
int num_err;
#define ERREUR_INCONNUE 0
#define ENTREE_REPERTOIRE_INCONNUE 1
#define DISQUE_INCONNU 2
#define DESCRIPTEUR_INCONNU 3
#define MAUVAISE_POSITION 4
#define PLUS_DE_NOEUD_LIBRE 5 
#define PLUS_DE_BLOC_LIBRE 6
#define TYPE_INCORRECT 7
#define CHEMIN_INCORRECT 8
#define RESTE_DES_LIENS 9
#define DISQUE_DEJA_MONTE 10
#define DISQUE_ERRONE 11
#define PLUS_DE_DEV_LIBRE 12
#define ENTREE_EXISTANTE 13
#define REF_INCORRECTE 14
#define DISQUE_COURANT 15
#define RESTE_DES_DISQUES_MONTES 16

/* Macro de positionnement (en octet) */
#define DEBUTNOEUD(n) (NOEUD2BLOC(n)*TAILLE_BLOC+NOEUDINBLOC(n)*sizeof(struct noeud))
#define DEBUTBLOC(n) ((TAILLE_ZONE_DESCRIPTION+TAILLE_ZONE_NOEUDS+(n))*TAILLE_BLOC)

/* Table des descripteurs */
#define MAX_OPEN_FILE 20 /* Nb max d'entrees ds la tables des descripteurs */
#define DESC_LIBRE -1
typedef struct { /* Structure de description des fichiers ouverts ds la table */
  desc_t desc;
  int mode;       /* Mode d'ouverture du fichier */
  short node;     /* Numero du noeud du fichier */
  short dev;      /* Numero du disque du fichier */
  void* buffer;   /* Tampon contenant un bloc du fichier */
  pos_t cur_pos;  /* Position courante dans le tampon */
  short cur_bloc; /* Numero du bloc charge dans le tampon */
  stat_p stat;    /* Informations sur le noeud du fichier */
} DESC;
DESC desc_table[MAX_OPEN_FILE]; /* Table des descripteurs */

/* Structures permettant de manipuler des noeuds et des blocs de
   disques montes differents */
struct numero_de_noeud {
  short num;
  short dev;
};
typedef struct numero_de_noeud n_noeud_s;
typedef struct numero_de_noeud *n_noeud_p;
struct numero_de_bloc {
  short num;
  short dev;
};
typedef struct numero_de_bloc n_bloc_s;
typedef struct numero_de_bloc *n_bloc_p;

/* Table des disques montes et structures associees */
#define MAX_FREE_NODES (((8*sizeof(int))/sizeof(short))-1) /* Evite les erreurs de bus */
#define MAX_FREE_BLOCS (TAILLE_BLOC/sizeof(short)-1)
#define MAX_MOUNT_DISK 10
#define DEV_LIBRE -1
typedef struct {
  short dev;        /* Numero du disque monte. Indique aussi si ce numero est libre */
  char* ref;        /* Nom de la racine du disque monte dans le disque principal */
  n_noeud_s pere;   /* Numero de noeud et de disque du rep pere */
  int desc;         /* descripteur du fichier contenant le disque */
  ino_t st_ino;     /* inoeud du fichier contenant le disque */
  dev_t st_dev;     /* id du peripherique du fichier contenant le disque */
  short *free_nodes_table; /* Table des blocs libres */
  short cur_p_node; /* Pointeur courant. Celui-ci est range ds la derniere case du
		       tableau free_nodes_table */
  short free_blocs_table[MAX_FREE_BLOCS+1]; /* Table des blocs libres */
  short cur_p_bloc; /* Pointeur courant. Celui-ci est range ds la derniere case du
		       tableau free_blocs_table */
} DEV;
DEV mount_table[MAX_MOUNT_DISK];
/* Num de descripteur du rep courant */
desc_t desc_cur_dir;

/* Noeud de la racine */
#define ROOT_NODE 0
#define ROOT_DEV 0

#endif
